// MKLINK.CPP - Written 2006 by Christoph H. Hochsttter (Christoph@Hochstaetter.org)

#include "mklink.h"


// Globale Variablen

String            RootPathName(MaxPath);        //Root-Pfadname des eingegebenen Links
String            VolumeName(64);               //Volume Label
DWORD             VolumeSerialNumber;           //Seriennumer
DWORD             MaximumComponentLength;       //Maximale Lnge einer Datei
DWORD             FileSystemFlags;              //Features vom Dateisystem
String            FileSystemName(64);           //Name des Dateisystems FAT, FAT, NTFS, UDF, CDFS, usw.
OSVERSIONINFOA    ovi;                          //Info ber Betriebssystemversion
String            fnTarget(MaxPath+4);          //Pfad des Ziels
String            fnLink(MaxPath);              //Pfad des Links
DWORD             aTarget,aLink;                //Attribute (hidden, read-only, etc. ) von Ziel und Link
UINT              gdt;                          //Art des Laufwerks (Netzwerk, Removable, CD, etc.)
int               copts                   = 0;  //Anzahl der Optionen
TCHAR             opt                     = 0;  //Welche Option /D /L /S etc.
HANDLE            hHeap                   = 0;  //Process Heap
DWORD             Version;                      //Versionsnummer, die auch mit NT 3.1 ausgelesen werden kann
HANDLE            hProcess;                     //Pseudo-Handle auf den eigenen Prozess (immer -1)
CPINFO            oemcp;                        //Info ber die aktuelle OEM Codepage


/*
Mit diesen Klassen werden lngere Strings auf
dem Heap angelegt
*/

const BOOL PHeapFree(const LPVOID mem)
{
  return HeapFree(hHeap, 0, mem);
}

const LPVOID PHeapAlloc(const size_t size)
{
  if ( !hHeap ) hHeap=GetProcessHeap();
  return HeapAlloc(hHeap, 0, size);
}

const LPVOID PHeapReAlloc(const LPVOID mem, const size_t size)
{
  return HeapReAlloc(hHeap, 0, mem, size);
}

StringW::StringW(const int size)
{
  buff = (LPWSTR)PHeapAlloc(size*sizeof(WCHAR));
  max = size;
}

StringW::~StringW()
{
  if (!PHeapFree(buff)) Error();
}

const BOOL StringW::Compact()
{
  if FAILED(StringCchLengthW(buff, max, &max)) return false;
  
  max++;
  buff = (LPWSTR)PHeapReAlloc(buff, max*sizeof(WCHAR));

  if (!buff) return false;
  return true;
}

StringA::StringA(const int size)
{
  buff = (LPSTR)PHeapAlloc(size*sizeof(char));
  max=size;
}

StringA::~StringA()
{
  if (!PHeapFree(buff)) Error();
}

const BOOL StringA::Compact()
{
  if FAILED(StringCchLengthA(buff, max, &max)) return false;

  max++;
  buff = (LPSTR)PHeapReAlloc(buff, max*sizeof(char));

  if (!buff) return false;
  return true;
}


/*
   Einige Helperfunktionen, um auf einer
   Console in der korrekten Codepage auszugeben
*/

#define MaxIob 3 * 1024

const int _tvofprintf(FILE* const f, const wchar_t *fmt, const va_list args)
{
  StringW temp(MaxIob);
  if (!temp) return -1;

  StringA oem(MaxIob*oemcp.MaxCharSize);
  if (!oem) return -1;

  register int i = vswprintf_s(temp, MaxIob, fmt, args);
  if ( i < 0) return i;

  CharToOemW(temp, oem);

  return fputs(oem, f);
}

const int _tvofprintf(FILE* const f, const char* const fmt, const va_list args)
{
  StringA oem(MaxIob*oemcp.MaxCharSize);
  if (!oem) return -1;

  register int i = vsprintf_s(oem, MaxIob, fmt, args);
  if ( i < 0 ) return i;

  CharToOemA(oem, oem);

  return fputs(oem, f);
}

const int _tofprintf(FILE* const f, const wchar_t* const fmt, ... )
{
  va_list args;

  va_start(args, fmt);
  const register int i=_tvofprintf(f, fmt, args);
  va_end(args);

  return i;
}

const int _tofprintf(FILE* const f, const char* const fmt, ... )
{
  va_list args;

  va_start(args, fmt);
  const register int i=_tvofprintf(f, fmt, args);
  va_end(args);

  return i;
}

const int __inline _tvoprintf(const wchar_t * const fmt, const va_list args)
{
  return _tvofprintf(stdout, fmt, args);
}

const int __inline _tvoprintf(const char* const fmt, va_list args)
{
  return _tvofprintf(stdout, fmt, args);
}

const int _toprintf(const wchar_t* const fmt, ... )
{
  va_list args;

  va_start(args, fmt);
  const register int i=_tvoprintf(fmt, args);
  va_end(args);

  return i;
}

const int _toprintf(const char* const fmt, ...)
{
  va_list args;

  va_start(args, fmt);
  const register int i=_tvoprintf(fmt, args);
  va_end(args);

  return i;
}


/*
Falls uns Windows einen Fehler meldet, wird der hier
ausgegeben und das Programm beendet.
*/

const void Error()
{
  DWORD errnum = GetLastError();
  LPTSTR errtext;

  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|
                FORMAT_MESSAGE_FROM_SYSTEM|
                FORMAT_MESSAGE_IGNORE_INSERTS,
                NULL, errnum, LANG_USER_DEFAULT, (LPTSTR) &errtext, 0, NULL);

  _tofprintf(stderr, _T("Error %lu: %s"), errnum, errtext);

  LocalFree(errtext); // Nur zum Spass, da wir das Programm beenden
  ExitProcess(1);
}


// Destructor von Junction muss nur den Heap wieder freigeben

FSLink::~FSLink()
{
  if (buf)
  {
    PHeapFree(buf);
  }
}


/*
Legt einen Buffer fr die REPARSE_DATA_BUFFER Struktur an.
Da der 16k gro ist, machen wir das lieber im Heap.
*/

const BYTE* const FSLink::Getbuf()
{
  Valid=false;

  buf = (BYTE*) PHeapAlloc(MAXIMUM_REPARSE_DATA_BUFFER_SIZE);
  Valid = (buf != NULL);  //Itanium macht Probleme mit "Valid = (BOOL)buf;"

  return buf;
}


/*
Initialisierung einer REPARSE_DATA_BUFFER Struktur
fr Junctions. Geht nur mit UNICODE.
*/

void FSLink::Private_Init(LPCWSTR const wLink, const ULONG ReparseTag, LPCWSTR wPrintName, const ULONG dwfLink)
{
  if (!Getbuf()) return;
  Valid=false;
  SetLastError(ERROR_BAD_PATHNAME);

  size_t size,sizePN;


  //Wir brauchen die Lnge des Strings in Bytes, nicht TCHARs
  
  if (StringCbLengthW(wLink, MaxPath, &size) != S_OK) return;
  if (!wPrintName) {
    wPrintName=wLink;
    if (!(dwfLink&LINK_ATTRIBUTE_RELATIVE)) wPrintName+=4;
  }
  if (StringCbLengthW(wPrintName, MaxPath, &sizePN) != S_OK) return;



  //Initialisierung der Struktur

  const register PREPARSE_DATA_BUFFER lrdb=rdb();

  lrdb->ReparseTag                    = ReparseTag;
  lrdb->Reserved                      = 0;
  lrdb->AllLinks.SubstituteNameOffset = 0;
  lrdb->AllLinks.SubstituteNameLength = (USHORT)size;
  lrdb->AllLinks.PrintNameOffset      = (USHORT)size+sizeof(WCHAR);
  lrdb->AllLinks.PrintNameLength      = (USHORT)sizePN;

  switch (ReparseTag) {

    case IO_REPARSE_TAG_MOUNT_POINT:

      lrdb->ReparseDataLength  = (USHORT)((size+sizePN) + 12);
      if ( FAILED(StringCbCopyW(lrdb->MountPointReparseBuffer.PathBuffer, MaxPath, wLink)) ) return;

      if ( FAILED(StringCbCopyW(lrdb->MountPointReparseBuffer.PathBuffer+size/sizeof(WCHAR)+1,
                                MaxPath, wPrintName)) )
      {
        return;
      }

      break;

    case IO_REPARSE_TAG_SYMBOLIC_LINK:


      /*
         Ab Windows Vista drfen Symbolic Links nur mit entsprechenden
         Privilegien angelegt werden, diese mssen wir zunchst vom
         System anfordern.
      */

      if (ovi.dwMajorVersion >= 6)
		  {
        HANDLE            hToken;
        TOKEN_PRIVILEGES  tp;

  		  if (!OpenProcessToken(hProcess, TOKEN_ADJUST_PRIVILEGES, &hToken)) return;

		    tp.PrivilegeCount = 1;
	      tp.Privileges->Attributes=SE_PRIVILEGE_ENABLED;
	      
        /*
           Kein Errorchecking, wir lassen es im Zweifel darauf
           ankommen, ob er den Symbolic Link trotzdem anlegt,
           zum Beispiel, weil wir ihn auf einem Netzlaufwerk
           anlegen, dass lter als Vista ist.
        */
        if ( LookupPrivilegeValueA("", "SeCreateSymbolicLinkPrivilege", &(tp.Privileges->Luid)) )
        {
	        AdjustTokenPrivileges(hToken, FALSE, &tp, sizeof(TOKEN_PRIVILEGES), NULL, NULL);
        }

        if ( !CloseHandle(hToken) ) return;
      }

      lrdb->ReparseDataLength = (USHORT)(size + sizePN + 16);
      lrdb->SymbolicLinkReparseBuffer.Type=dwfLink;

      if FAILED(StringCbCopyW(lrdb->SymbolicLinkReparseBuffer.PathBuffer, MaxPath, wLink)) return;
      
      if FAILED(StringCbCopyW(lrdb->SymbolicLinkReparseBuffer.PathBuffer + size/sizeof(WCHAR)+1, MaxPath, wPrintName)) return;

      break;

    default:
      SetLastError(ERROR_REPARSE_TAG_INVALID);
      return;
  }

  Valid=true;
  SetLastError(ERROR_SUCCESS);
}



/*
ANSI Version des Constructors. Vor Init
muss der ANSI String in Unicode umgewandelt werden
*/

FSLink::FSLink(LPCSTR const Link,
               const ULONG ReparseTag,
               LPCSTR const PrintName,
               const ULONG dwfLink)
{
  
  size_t size,sizePN;
  Valid=false;
  SetLastError(ERROR_BAD_PATHNAME);

  if FAILED(StringCchLengthA(Link, MaxPath, &size)) return;
  ++size;

  StringW  wLink(MaxPath+4);
  if (!wLink) return;

  WCHAR* wPrintName;

  if ( !MultiByteToWideChar(CP_THREAD_ACP, MB_PRECOMPOSED, Link, (int)size, wLink, MaxPath+4) ) return;
  if ( !wLink.Compact() ) return;
  if (PrintName)
  {
    if ( StringCchLengthA(PrintName, MaxPath, &sizePN) != S_OK ) return;
    ++sizePN;
    wPrintName = (WCHAR*)PHeapAlloc((MaxPath+4)*sizeof(WCHAR));
    if ( !wPrintName ) return;

    if ( !MultiByteToWideChar(CP_THREAD_ACP, MB_PRECOMPOSED, PrintName, (int)sizePN, wPrintName, MaxPath+4) ) return;
  }
  else
  {
    wPrintName=0;
  }
  Private_Init(wLink, ReparseTag, wPrintName, dwfLink);
  if (wPrintName) PHeapFree(wPrintName);

  SetLastError(ERROR_SUCCESS);
  return;
}



/*
Konstruktion der Reparse_Point Klasse. Es wird das
Handle auf ein Directory ermittelt. Um ein Directory zu ffnen
hat sich Microsoft berlegt, dass man bei CreateFile
FILE_FLAG_BACKUP_SEMANTICS setzen muss. Nicht besonders logisch,
ist aber so dokumentiert.
*/

Reparse_Point::Reparse_Point(LPCTSTR const szDir, const BOOL Writeable)
  : hLink(CreateFile(szDir,
                     GENERIC_READ|(Writeable ? GENERIC_WRITE : 0),
                     FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
                     NULL,
                     OPEN_ALWAYS,
                     FILE_FLAG_BACKUP_SEMANTICS|FILE_FLAG_OPEN_REPARSE_POINT,
                     0)) {}


/*
Der Destructor schliet einfach nur das Handle wieder
*/

Reparse_Point::~Reparse_Point()
{
  if (hLink != INVALID_HANDLE_VALUE) CloseHandle(hLink);
}


/*
Um einen Reparse Point zu setzen, wird dies einfach
mit DeviceIoControl und FSCTL_SET_REPARSE_POINT erledigt.
*/

const BOOL Reparse_Point::Set(const REPARSE_DATA_BUFFER* const rdb)
{
  if (!IsValid())
  {
    SetLastError(ERROR_BAD_PATHNAME);
    return false;
  }

  DWORD dummy;

  return DeviceIoControl(hLink, FSCTL_SET_REPARSE_POINT,
                        (LPVOID)rdb, rdb->ReparseDataLength+REPARSE_DATA_BUFFER_HEADER_SIZE,
                        NULL, 0, &dummy, 0);
}


const BOOL CreateFSLink( LPCTSTR const Link,
                         LPCTSTR const Target,
                         const ULONG ReparseTag,
                         LPCTSTR const PrintName,
                         const ULONG dwfLink)

{

  //In temp schreiben wir den endgltigen Pfadnamen

  String temp(MaxPath+4);
  if (!temp) return false;

  SetLastError(ERROR_BAD_PATHNAME);

  /*
  Absolute Link beginnen mit \??\, um zu kennzeichnen,
  dass es IMMER ein ABSOLUTER Pfad ist.

  */

  if ( (dwfLink & LINK_ATTRIBUTE_RELATIVE) || !(_tmemcmp(_T("\\?"), Target, 2)) )
  {
    if (StringCchCopy(temp, MaxPath+4, Target) != S_OK) return false;
  }
  else
  {
    int UncOffset = 0;
    if (StringCchCopy(temp, MaxPath+4, _T("\\??\\")) != S_OK) return false;
    if ( !_tmemcmp(_T("\\\\"), Target, 2) )
    {
      if (StringCchCat(temp, MaxPath+4, _T("UNC")) != S_OK) return false;
      UncOffset = 1;
    }
    String temp2(MaxPath);
    if (!temp2) return false;
    if (!GetFullPathName(Target, MaxPath, temp2, NULL)) return false;
    if (StringCchCat(temp, MaxPath, temp2+UncOffset) != S_OK) return false;
  }
  
  if ( !temp.Compact() ) return false;
  FSLink j(temp,ReparseTag,PrintName,dwfLink);

  if (!j.IsValid() || !Reparse_Point(Link, true).Set(j.rdb())) return false;
  SetLastError(ERROR_SUCCESS);
  return true;
}


//Siehe Beschreibung in MKLINK.H

const DWORD GetFSLinkTarget(LPCTSTR const Link,
                            LPTSTR const PrintName,
                            const DWORD PrintNameMaxSize,
                            LPTSTR const SubstituteName,
                            const DWORD SubstituteNameMaxSize,
                            ULONG* const ReparseTag,
                            BOOL* const IsRelative)
{
  Reparse_Point Dir(Link);
  SetLastError(ERROR_PATH_NOT_FOUND);

  if (!Dir.IsValid()) return false;


  //Leeren Link erstellen

  FSLink j;
  DWORD dummy;


  //Link Target ermitteln
  
    const register REPARSE_DATA_BUFFER* rdb=j.rdb();
    if ( !DeviceIoControl(Dir,
                          FSCTL_GET_REPARSE_POINT,
                          NULL,
                          0,
                          (LPVOID)rdb,
                          MAXIMUM_REPARSE_DATA_BUFFER_SIZE,
                          &dummy,0) )
    {
      return false;
    }

    *ReparseTag = rdb->ReparseTag;


// Wir haben diese kranken Strukturen nicht erfunden
// Beschwerden an billg@microsoft.com

#ifdef UNICODE

  SetLastError(ERROR_INSUFFICIENT_BUFFER);
  switch(rdb->ReparseTag) {

    case IO_REPARSE_TAG_MOUNT_POINT:

      if ( StringCbCopyNW(PrintName, PrintNameMaxSize,
                          rdb->MountPointReparseBuffer.PathBuffer + rdb->MountPointReparseBuffer.PrintNameOffset / sizeof(WCHAR),
                          rdb->MountPointReparseBuffer.PrintNameLength
                          ) != S_OK
      ) return false;

      if ( StringCbCopyNW(SubstituteName, SubstituteNameMaxSize,
                          rdb->MountPointReparseBuffer.PathBuffer + rdb->MountPointReparseBuffer.SubstituteNameOffset / sizeof(WCHAR),
                          rdb->MountPointReparseBuffer.SubstituteNameLength
                          ) != S_OK
      ) return false;

      *IsRelative = false;
      break;

    case IO_REPARSE_TAG_SYMBOLIC_LINK:

      if ( StringCbCopyNW(PrintName, PrintNameMaxSize,
                          rdb->SymbolicLinkReparseBuffer.PathBuffer + rdb->SymbolicLinkReparseBuffer.PrintNameOffset / sizeof(WCHAR),
                          rdb->SymbolicLinkReparseBuffer.PrintNameLength
                          ) != S_OK
      ) return false;

      if ( StringCbCopyNW(SubstituteName, SubstituteNameMaxSize,
                          rdb->SymbolicLinkReparseBuffer.PathBuffer + rdb->SymbolicLinkReparseBuffer.SubstituteNameOffset / sizeof(WCHAR),
                          rdb->SymbolicLinkReparseBuffer.SubstituteNameLength
                          ) != S_OK
      ) return false;

      *IsRelative = (rdb->SymbolicLinkReparseBuffer.Type & 1);
      break;

    default:
      /* fputs("ERROR! Unsupported reparse point found.\n"
            "MKLINK supports only junctions and symbolic links.\n", stderr); */
      SetLastError(ERROR_REPARSE_TAG_INVALID);
      return false;
  }

#else

  switch(rdb->ReparseTag) {

    case IO_REPARSE_TAG_MOUNT_POINT:

      if ( !WideCharToMultiByte(CP_THREAD_ACP, 0,
                                rdb->MountPointReparseBuffer.PathBuffer + rdb->MountPointReparseBuffer.PrintNameOffset / sizeof(WCHAR),
                                rdb->MountPointReparseBuffer.PrintNameLength / sizeof(WCHAR) + 1,
                                PrintName, PrintNameMaxSize ,NULL , NULL
                                )
      ) return false;

      PrintName[ rdb->MountPointReparseBuffer.PrintNameLength / sizeof(WCHAR) ] = 0;

      if ( !WideCharToMultiByte(CP_THREAD_ACP, 0,
                                rdb->MountPointReparseBuffer.PathBuffer + rdb->MountPointReparseBuffer.SubstituteNameOffset / sizeof(WCHAR),
                                rdb->MountPointReparseBuffer.SubstituteNameLength / sizeof(WCHAR) + 1,
                                SubstituteName, SubstituteNameMaxSize ,NULL, NULL
                                )
      ) return false;

      SubstituteName[ rdb->MountPointReparseBuffer.SubstituteNameLength / sizeof(WCHAR) ] = 0;
      *IsRelative = false;
      break;

    case IO_REPARSE_TAG_SYMBOLIC_LINK:

      if ( !WideCharToMultiByte(CP_THREAD_ACP, 0,
                                rdb->SymbolicLinkReparseBuffer.PathBuffer + rdb->SymbolicLinkReparseBuffer.PrintNameOffset / sizeof(WCHAR),
                                rdb->SymbolicLinkReparseBuffer.PrintNameLength / sizeof(WCHAR) + 1,
                                PrintName, PrintNameMaxSize ,NULL , NULL
                                )
      ) return false;

      PrintName[ rdb->SymbolicLinkReparseBuffer.PrintNameLength / sizeof(WCHAR) ] = 0;

      if ( !WideCharToMultiByte(CP_THREAD_ACP, 0,
                                rdb->SymbolicLinkReparseBuffer.PathBuffer + rdb->SymbolicLinkReparseBuffer.SubstituteNameOffset / sizeof(WCHAR),
                                rdb->SymbolicLinkReparseBuffer.SubstituteNameLength / sizeof(WCHAR) + 1 ,
                                SubstituteName, SubstituteNameMaxSize ,NULL , NULL
                                )
      ) return false;
      
      SubstituteName[ rdb->SymbolicLinkReparseBuffer.SubstituteNameLength / sizeof(WCHAR) ] = 0;
      *IsRelative = (rdb->SymbolicLinkReparseBuffer.Type & 1);
      break;

    default:
      SetLastError(ERROR_REPARSE_TAG_INVALID);
      return false;
  }

#endif

  SetLastError(ERROR_SUCCESS);
  return true;
}


void ReparsePointError()
{
  if ( !(FileSystemFlags & FILE_SUPPORTS_REPARSE_POINTS) )
  {
    _tofprintf(stderr,
               _T("%s-Drive %s (%s) does not support reparse points.\n"),
               (LPCTSTR)FileSystemName,
               (LPCTSTR)RootPathName,
               (LPCTSTR)VolumeName);

    SetLastError(ERROR_NOT_A_REPARSE_POINT);
    Error();
  }
}


/*
Kleine Hilfe, falls Aufruf ohne Parameter
*/

void Syntax()
{
  static const char Message[] =

#ifdef _DEBUG
    "BETA VERSION DO NOT USE FOR PRODUCTION!!!!!!!!!!!!!!!!!!!!!!!!\n\n"
#endif
    "MKLINK 1.01 - (c) 2006 Christoph H. Hochsttter and CNET Networks.\n"
    "Freeware, can be used by anyone for any purpose without restrictions.\n"
    "Absolutely no warranty, neither expressed nor implied.\n\n"
		"Windows Vista like hardlink and junction utility.\n\n"
    "Usage:\n"
    "  MKLINK [/D] | [/H] | [/I] | [/J] | [/L] | [/S] <Link> [<target>]\n\n"
		"    /D Creates a symbolic link for a directory.\n"
		"    /H Creates a hard link for a file on the same partition.\n"
		"    /I Displays filesystem information for a given pathname.\n"
		"    /J Creates a junction for a directory.\n"
    "    /L Lists a junction.\n"
    "    /S Creates a symbolic link for a file.\n"
		"\n  Examples:\n"
    "    MKLINK /J C:\\Myfiles \042C:\\Documents and Settings\\Joe.Cool\\My Documents\042\n"
    "    MKLINK /H C:\\Myfiles\\Destination.txt C:\\Myfiles\\Source.txt\n\n"
    "  To remove a hard link or a file symbolic link use: DEL|ERASE <target>\n\n"
    "  To remove a junction or directory symbolic link use: RMDIR|RD <target>\n"
    "  RMDIR (without /S switch) will not delete the original directory.\n";

  _toprintf(Message);
}


void DisplayLinkInfo(const TCHAR * const argv[])
{
      String PrintName(MaxPath+4);
      if (!PrintName) Error();

      String SubstituteName(MaxPath+4);
      if (!SubstituteName) Error();
      ULONG  ReparseTag;
      BOOL   IsRelative;

      if ( !GetFSLinkTarget(fnLink, PrintName, MaxPath+4, SubstituteName, MaxPath+4, &ReparseTag, &IsRelative) )
      {
        ReparsePointError();
        Error();
      }

      if ( !SubstituteName.Compact() || !PrintName.Compact() ) Error();

      switch( ReparseTag ) {

        case IO_REPARSE_TAG_MOUNT_POINT:
        case IO_REPARSE_TAG_SYMBOLIC_LINK:
          _toprintf(_T("\nLink target         : %s -> %s\n"), argv[1+copts], (LPCTSTR)SubstituteName+(IsRelative?0:4));
          printf("Target refers to    : ");
          if (IsRelative) {
            puts("All computers, since link is relative");
          }
          else
          {
            if ( (gdt != DRIVE_REMOTE) || (ReparseTag == IO_REPARSE_TAG_SYMBOLIC_LINK) ) {
              puts("Namespace on this local computer");
            }
            else
            {
              puts("Namespace on network server, where it resides");
            }
          }
          if (*PrintName) {
            _toprintf(_T("Display target      : %s -> %s\n"), argv[1+copts], (LPCTSTR)PrintName);
          }
          printf("Type of link        : ");
          if (aLink & FILE_ATTRIBUTE_DIRECTORY) {
            printf("Directory ");
          }
          else
          {
            printf("File ");
          }
          puts(((ReparseTag==IO_REPARSE_TAG_MOUNT_POINT)?"junction":"symbolic link"));


          break;
        default:
          printf("Unknown reparse point");
      }
}

const int __cdecl _tmain(const int argc, const TCHAR * const argv[])
{

  /*
  Als erstes die Version checken, immer mit der ASCII
  Variante des API, damit wir unter Win95/98/Me wenigstens
  eine Fehlermeldung bekommen.
  */

  Version=GetVersion();

  if ((Version&255) < 5)
  {
    fputs("Windows 2000 or later required (e.g. 2000/2003/XP)!\n", stderr);
    return -1;
  }

  ovi.dwOSVersionInfoSize=sizeof(OSVERSIONINFOA);
  GetVersionExA(&ovi);

  hProcess=GetCurrentProcess(); //Gibt immmer -1 zurck

  int cfiles = argc;

  if (!GetCPInfo(CP_OEMCP,&oemcp)) Error();


  // Aufruf ohne Parameter -> Hilfe anzeigen

  if (argc == 1)
  {
    Syntax();
    return 0;
  }


  /* Aufruf mit oder ohne Option */

  if ((argv[1][0] == '/' || argv[1][0] == '-'))
  {
    --cfiles;
    ++copts;
  }

  if (cfiles < 2)
  {
    fputs("At least one filename is required.\n", stderr);
    return -1;
  }

  if (cfiles > 4)
  {
    fputs("A maximum of 3 arguments can be specified.\n", stderr);
    return -1;
  }

  if ((argv[1+copts][0] == '/' || argv[1+copts][0] == '-'))
  {
    fputs("Only one option can be specified.\n", stderr);
    return -1;
  }

  if (!GetFullPathName(argv[1+copts], MaxPath, fnLink, NULL)) Error();
  if ( !fnLink.Compact() ) Error();

  aLink=GetFileAttributes(fnLink);
  
  SetErrorMode(SEM_NOOPENFILEERRORBOX|SEM_FAILCRITICALERRORS);

  if ( !GetVolumePathName(fnLink, RootPathName, MaxPath) ) Error();
  if ( !RootPathName.Compact() ) Error();

  if ( !GetVolumeInformation(RootPathName,
                             VolumeName, 64,
                             &VolumeSerialNumber,
                             &MaximumComponentLength,
                             &FileSystemFlags,
                             FileSystemName, 64))
  {
    Error();
  }

  if ( !VolumeName.Compact() || !FileSystemName.Compact() ) Error();

  gdt=GetDriveType(RootPathName);

  if ( cfiles != 2 )
  {
    ReparsePointError();
  }

  if (cfiles == 2 && aLink == INVALID_FILE_ATTRIBUTES)
  {
    _tofprintf(stderr, _T("File %s\n"), argv[1+copts]);
    SetLastError(ERROR_FILE_NOT_FOUND);
    Error();
  }

  if (cfiles > 2)
  {
    if ( !_tmemcmp(_T("\\?"),argv[2+copts],2) )
    {
      if FAILED(StringCchCopy(fnTarget,MaxPath+4,argv[2+copts])) Error();
    }
    else
    {
      if ( !GetFullPathName(argv[2+copts], MaxPath, fnTarget, NULL) ) Error();
      if ( !fnTarget.Compact() ) Error();
      aTarget=GetFileAttributes(fnTarget);
    }
  }


  /*
  Wenn keine Option angegeben wurde, dann mssen wir eben "raten", was
  der Benutzer will:

  Regel 1: Wenn nur ein Parameter angegeben wurde, dann /L annehmen
  Regel 2: Unter Vista, wenn Dateisystem Symbolic Links untersttzt /S annehmen
  Regel 3: Wenn Ziel ein Verzeichnis ist, dann /J annehmen
  Regel 4: Wenn Ziel eine Datei ist, dann /H annehmen.

  Sobald eine Regel greift, werden die anderen nicht mehr abgearbeitet.
  */

  if (!copts)
  {
    switch(cfiles)
    {
      case 2:
        opt = 'L';
        break;
      case 3:
      case 4:
        if ( (ovi.dwMajorVersion > 5) && (FileSystemFlags&FILE_SUPPORTS_TRANSACTIONS) )
        {
          opt = 'S';
        }
        else {
          opt = 'H';
          if ( aTarget & FILE_ATTRIBUTE_DIRECTORY ) opt = 'J';
        }
        break;
    }
  }
  else
  {
    opt=argv[1][1];
  }

  /*
  Jetzt fhren wir die eigentlichen Aktionen aus.
  */

  switch(opt|0x20)
  {
    case 'i':
      if (cfiles != 2) {
        fputs("Option /I requires exactly one argument.\n", stderr);
        return -1;
      }

      printf("\nType of Drive    : ");

      if (gdt == DRIVE_UNKNOWN)   printf("UNKNOWN");
      if (gdt == DRIVE_REMOVABLE) printf("REMOVABLE MEDIA");
      if (gdt == DRIVE_FIXED)     printf("NON-REMOVABLE MEDIA");
      if (gdt == DRIVE_REMOTE)    printf("NETWORK");
      if (gdt == DRIVE_CDROM)     printf("CD-/DVD-ROM");
      if (gdt == DRIVE_RAMDISK)   printf("RAMDISK");

      putchar(0xa);

      _toprintf(_T("Root path        : %s\n"),(LPCTSTR)RootPathName);
      _toprintf(_T("Filesystem       : %s\n"),(LPCTSTR)FileSystemName);
      _toprintf(_T("Volume label     : %s"),(LPCTSTR)VolumeName);

      if (VolumeSerialNumber) {
        _toprintf(_T(" (%04X-%04X)"),HIWORD(VolumeSerialNumber),LOWORD(VolumeSerialNumber));
      }

      putchar(0xa);

      printf("Features         :");

      if (FileSystemFlags & FILE_SUPPORTS_ENCRYPTION)     printf(" ENCRYPTION");
      if (FileSystemFlags & FILE_FILE_COMPRESSION)        printf(" FILECOMPRESSION");
      if (FileSystemFlags & FILE_VOLUME_QUOTAS)           printf(" QUOTA"); 
      if (FileSystemFlags & FILE_READ_ONLY_VOLUME)        printf(" READONLY");
      if (FileSystemFlags & FILE_SUPPORTS_REPARSE_POINTS) printf(" REPARSEPOINTS");
      if (FileSystemFlags & FILE_PERSISTENT_ACLS)         printf(" SECURITY");
      if (FileSystemFlags & FILE_NAMED_STREAMS)           printf(" STREAMS");
      if (FileSystemFlags & FILE_SUPPORTS_TRANSACTIONS)   printf(" TRANSACTIONS");
      if (FileSystemFlags & FILE_UNICODE_ON_DISK)         printf(" UNICODE"); 
      if (FileSystemFlags & FILE_VOLUME_IS_COMPRESSED)    printf(" VOLUMECOMPRESSION");

      puts("\n");

      if (FileSystemFlags & FILE_SUPPORTS_REPARSE_POINTS) {
        puts("This drive supports JUNCTIONS.");
      }
      else
      {
        puts("This drive does NOT support JUNCTIONS.");
      }

      if (FileSystemFlags & FILE_SUPPORTS_TRANSACTIONS)  //Kleiner Hack
      {
        printf("This drive supports SYMBOLIC LINKS");
        if ( ovi.dwMajorVersion < 6 )
        {
          printf(", but you cannot use them on this computer");
        }
        puts(".");
      }
      else
      {
        puts("This drive does NOT support SYMBOLIC LINKS.");
      }

      if (gdt == DRIVE_REMOTE) {
        fputs("\nWARNING! A network drive possibly does not return the correct root path.\n", stderr);
      }

      break;

    case 'l':
      if (cfiles > 2)
      {
        fputs("Only one link can be specified with option /L.\n", stderr);
        return -1;
      }
      DisplayLinkInfo(argv);
      break;

    case 'h':
      if (cfiles != 3)
      {
        fputs("Option /H requires two arguments.\n", stderr);
        return -1;
      }

      if (!CreateHardLink(fnLink, fnTarget, NULL))
      {
        _tofprintf(stderr, _T("Hardlink %s <-> %s\n"), argv[1+copts], argv[2+copts]);
        Error();
      }

      _toprintf(_T("Created Hardlink %s <-> %s\n"), argv[1+copts], argv[2+copts]);
      break;

    case 'j':
    case 'd':
    case 's':
      DWORD dwMkDir;
      dwMkDir=0;

      if (cfiles < 3)
      {
        fputs("Options /D, /J and /S require at least two arguments.\n", stderr);
        return -1;
      }


      /*
      Junctions und Symbolische Links mssen in einem leeren Verzeichnis erstellt werden.
      Wenn das Verzeichnis nicht existiert, dann mssen wir es eben
      anlegen.
      */

      if ( aLink == INVALID_FILE_ATTRIBUTES && ((opt&0xDF) == 'D' || (opt&0xDF) == 'J') )
      {
        dwMkDir = CreateDirectory(fnLink, NULL);
        if (!dwMkDir)
        {
          _tofprintf(stderr, _T("Directory %s\n"), argv[1+copts]);
          Error();
        }
      }

      register BOOL IsRel,IsJunction;
      IsJunction = ( (opt & 0xDF) == 'J');
      IsRel = ((opt & 0xDF) != 'J') && _tmemcmp(argv[2+copts],_T("\\\\"),2) && _tmemcmp(argv[2+copts],_T("\\?"),2) && (argv[2+copts][1] != ':');

      if ( !CreateFSLink(fnLink,
                         IsRel ? argv[2+copts] : fnTarget,
                         IsJunction ? IO_REPARSE_TAG_MOUNT_POINT : IO_REPARSE_TAG_SYMBOLIC_LINK,
                         (cfiles == 3) ? NULL : argv[3+copts],
                         IsRel ? LINK_ATTRIBUTE_RELATIVE : 0) )

	    {
        _tofprintf(stderr,_T("Could not create link: %s -> [%s].\n\n"), argv[1+copts], argv[2+copts]);
	  
        if (dwMkDir && !RemoveDirectory(fnLink))
        {
          _tofprintf(stderr,_T("WARNING! MKLINK created Directory %s for a reparse point,\n")
                            _T("but was unable to remove it, after a reparse point creation failure.\n"), argv[1+copts]);
        }
		  Error();
      }

      _toprintf(_T("Created link %s -> [%s].\n"), argv[1+copts], (LPCTSTR)fnTarget);
      aLink=GetFileAttributes(fnLink);

      if (!IsJunction)
      {
        if (ovi.dwMajorVersion<6)
        {
          fputs("\nWARNING! The symbolic link can not be used on this computer,\n"
                "because it does not run Windows Vista or newer.", stderr);
          if (gdt != DRIVE_REMOTE)
          {
            fputs(" However, if\n"
                  "you upgrade your computer or will use this volume on Vista\n"
                  "or a newer OS, this symbolic link will become available.", stderr);
          }
          fputs("\n\n", stderr);
        }
        if ( gdt==DRIVE_REMOTE && !(FileSystemFlags&FILE_SUPPORTS_TRANSACTIONS))
        {
          fputs("\nWARNING! You created a symbolic link on a network computer, that\n"
                "does not run Windows Vista or newer. This computer cannot make use\n"
                "of the symbolic link and cannot make it available to other computers.\n\n", stderr);
        }
      }
      DisplayLinkInfo( argv );
      break;

    default:
      fputs("Unknown option specified. Only /D /H /I /J /L /S and are valid.\n", stderr);
      return -1;
  }
  return 0;
}

