//MKLINK.H - Written 2006 by Christoph H. Hochsttter (Christoph@Hochstaetter.org)

#define _WIN32_WINNT 0x500

#include <windows.h>
#include <winioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <strsafe.h>


/*
Ein paar TCHAR Funktionen, die Microsoft im SDK vergessen hat
*/

#ifdef _UNICODE
  #define _tfputs fputws
  #define _tfputc fputwc
  #define _tfprintf fwprintf
  #define _tsscanf swscanf
  #define _tmemcmp wmemcmp
  #define _tmemcpy wmemcpy
#else
  #define _tfputs fputs
  #define _tfputc fputc
  #define _tfprintf fprintf
  #define _tsscanf sscanf
  #define _tmemcmp memcmp
  #define _tmemcpy memcpy
#endif

#define LINK_ATTRIBUTE_RELATIVE 1 // Hab ich erfunden, da keine Doku verfgbar

/*
Microsoft hat diese struct aus dem SDK entfernt, obwohl
sie dort explizit erwhnt wird. (Offensichtlich will man nicht,
dass mit Junctions programmiert wird. Man findet sie aber unter
http://msdn.microsoft.com/library/default.asp?url=/library/en-us/ifsk_r/hh/IFSK_r/fileinformationstructures_4f1b658e-1833-421f-a726-448b20b1c595.xml.asp.
Ohne selbige kann man mit Reparse Points wie Junctions gar
nichts programmieren.
*/

#ifndef _REPARSE_DATA_BUFFER 
typedef struct _REPARSE_DATA_BUFFER {
  ULONG  ReparseTag;
  USHORT  ReparseDataLength;
  USHORT  Reserved;
  union {
    struct {
      USHORT  SubstituteNameOffset;
      USHORT  SubstituteNameLength;
      USHORT  PrintNameOffset;
      USHORT  PrintNameLength;
    } AllLinks;
    struct {
      USHORT  SubstituteNameOffset;
      USHORT  SubstituteNameLength;
      USHORT  PrintNameOffset;
      USHORT  PrintNameLength;
      ULONG   Type;  //Vermute ich mal, ist nicht dokumentiert (1=relativer Link)
      WCHAR   PathBuffer[MAXIMUM_REPARSE_DATA_BUFFER_SIZE-20];
      } SymbolicLinkReparseBuffer;
    struct {
      USHORT  SubstituteNameOffset;
      USHORT  SubstituteNameLength;
      USHORT  PrintNameOffset;
      USHORT  PrintNameLength;
      WCHAR   PathBuffer[MAXIMUM_REPARSE_DATA_BUFFER_SIZE-16];
      } MountPointReparseBuffer;
    struct {
      UCHAR  DataBuffer[MAXIMUM_REPARSE_DATA_BUFFER_SIZE-8];
    } GenericReparseBuffer;
  };
} REPARSE_DATA_BUFFER, *PREPARSE_DATA_BUFFER;
#endif

#ifndef REPARSE_DATA_BUFFER_HEADER_SIZE
#define REPARSE_DATA_BUFFER_HEADER_SIZE FIELD_OFFSET(REPARSE_DATA_BUFFER, GenericReparseBuffer)
#endif

#ifndef IO_REPARSE_TAG_SYMBOLIC_LINK
#define IO_REPARSE_TAG_SYMBOLIC_LINK 0xA000000C //Aus dem Vista Filesystem rausgehackt, Name erfunden
#endif

#ifndef FILE_SUPPORTS_TRANSACTIONS
#define FILE_SUPPORTS_TRANSACTIONS 0x00200000   // Aus der Vista SDK-Doku geklaut
#endif


//Pfadlngen bis zum halben ReparseBuffer untersttzen minus Prefix fr \??\

#ifdef _UNICODE
#define MaxPath MAXIMUM_REPARSE_DATA_BUFFER_SIZE/2-FIELD_OFFSET(REPARSE_DATA_BUFFER, SymbolicLinkReparseBuffer.PathBuffer)/2-8
#else
#define MaxPath _MAX_PATH
#endif

//Klasse fr Filesystem Links.

class FSLink
{
  public:

    REPARSE_DATA_BUFFER* rdb() const {return (REPARSE_DATA_BUFFER*)buf; }

    const BOOL IsValid() const { return Valid; }

    FSLink(LPCSTR const Link,
           const ULONG ReparseTag = IO_REPARSE_TAG_MOUNT_POINT,
           LPCSTR const PrintName = NULL,
           const ULONG dwfLink = 0
           );     //ANSI

    FSLink(LPCWSTR const wLink,
           ULONG ReparseTag = IO_REPARSE_TAG_MOUNT_POINT,
           LPCWSTR wPrintName = NULL,
           const ULONG dwfLink = 0
           )     //UNICODE
    
    {
      Private_Init(wLink, ReparseTag, wPrintName, dwfLink);
    }

    FSLink() { Getbuf(); }

    ~FSLink();

  private:
    BYTE*                buf;
    BOOL                 Valid;

    void                 Private_Init(LPCWSTR const wLink,
                                      ULONG ReparseTag = IO_REPARSE_TAG_MOUNT_POINT,
                                      LPCWSTR wPrintName = NULL,
                                      const ULONG dwfLink = 0);

    const BYTE* const    Getbuf();
};


/* Klasse um bei Directories Reparse Points zu setzen */

class Reparse_Point
{
public:
	            Reparse_Point(LPCTSTR const path,
                            const BOOL ReadWrite = false);

	            ~Reparse_Point();

	const BOOL  Set(const REPARSE_DATA_BUFFER* const);

	operator    HANDLE() const   { return hLink; }
	const BOOL  IsValid() const  { return hLink != INVALID_HANDLE_VALUE; }

private:
	HANDLE    hLink;
};


/* Klassen fr Strings mit Heap Allocation */

class StringW
{
  public:
                StringW(const int size);
                ~StringW();
    operator     LPWSTR() const             { return buff; }
    const BOOL   Compact();
    const size_t GetMax() const             { return max; }

  private:
    LPWSTR      buff;
    size_t      max;
};


class StringA
{
  public:
                 StringA(const int size);
                 ~StringA();
    operator     LPSTR() const              { return buff; }
    const BOOL   Compact();
    const size_t GetMax() const             { return max; }

  private:
    LPSTR       buff;
    size_t      max;
};

#ifdef _UNICODE
#define String StringW
#else
#define String StringA
#endif


/*
Einfach zu nutzende Funktionen um einen Filesystem Link zu
erstellen oder das Target auszulesen.
*/

const BOOL CreateFSLink(LPCTSTR const Link,              // IN  Pfad, der in einen Link umgewandelt werden soll
                        LPCTSTR const Target,            // IN  Darauf soll der Link zeigen
                        const ULONG ReparseTag,          // IN  Art des Links z.B. IO_REPARSE_TAG_MOUNT_POINT
                        LPCTSTR const PrintName = NULL,  // IN  Diese Info wird dem Benutzer ab Vista angezeigt, kann NULL sein
                        const ULONG dwfLink = 0);        // IN  Linkspezifische Flags


const DWORD GetFSLinkTarget(LPCTSTR const Link,                  // IN  Pfad des Reparse Points
                            LPTSTR const PrintName,              // OUT Pfad der Vista Info, gibt "" zurck, wenn nicht vorhanden
                            const DWORD PrintNameMaxSize,        // IN  Gre des Buffers
                            LPTSTR const SubstituteName,         // OUT Ziel des Links
                            const DWORD SubstituteNameMaxSize,   // IN  Gre des Buffers
                            ULONG* const ReparseTag,             // OUT Art des Links z.B. IO_REPARSE_TAG_MOUNT_POINT
                            BOOL* const IsRelative);             // OUT Relatives oder absolutes Ziel des Links


const void Error();


/*
Funktionen um auf der Console in der richtige Codepage zu schreiben
*/

const int _tvofprintf(FILE* const f, const wchar_t *fmt, const va_list args);  // UNICODE
const int _tvofprintf(FILE* const f, const char* const fmt, va_list args);     // ANSI

const int __inline _tvoprintf(const wchar_t * const fmt, const va_list args);  // UNICODE
const int __inline _tvoprintf(const char* const fmt, const va_list args);      // ANSI

const int _tofprintf(FILE* const f, const wchar_t* const fmt, ... );           // UNICODE
const int _tofprintf(FILE* const f, const char* const fmt, ... );              // ANSI

const int _toprintf(const wchar_t* const fmt, ... );                           // UNICODE
const int _toprintf(const char* const fmt, ...);                               // ANSI

